-- kissWxEuphoria.e
-- Keep it Simple Stupid, WxEuphoria
-- Jean-Marc Duro
-- 2016-01-02
-- v0.0.2

include std/search.e
include std/filesys.e
include std/convert.e
include std/math.e
include std/machine.e
include std/dll.e
public include wxeud.e as wx
include myLibs/myDebug.e
include myLibs/common.e
include myLibs/myFile.e

public constant  -- used by gridSetup
  wxSTRING=0, wxINTEGER=1, wxFLOAT=2, wxBOOL=3

constant  -- used by gridSetup
  FIELD_NAME=1, FIELD_SIZE=2, FIELD_TYPE=3, FIELD_ALIGNMENT=4, FIELD_WIDTH=5, FIELD_PRECISION=6

-- stores object ids with their class
sequence registry = {}

constant REG_ID=1, REG_CLASS=2, REG_CHILDREN=3, REG_PARAMS=4, REG_TIP=5
public constant TIP_NONE=0, TIP_AUTO=1
public constant wxROW=1, wxCOL=2, wxCELL=3
public constant wxHEADER=1, wxFOOTER=2

constant WX_EUPHORIA_CLASSES = {
  {wxAboutDialogInfo, "wxAboutDialogInfo"},
--  {wxApp, "wxApp"},
  {wxArrayStringProperty, "wxArrayStringProperty"},
  {wxAuiDockArt, "wxAuiDockArt"},
  {wxAuiManager, "wxAuiManager"},
  {wxAuiNotebook, "wxAuiNotebook"},
  {wxAuiPaneInfo, "wxAuiPaneInfo"},
  {wxBitmap, "wxBitmap"},
  {wxBitmapButton, "wxBitmapButton"},
  {wxBitmapComboBox, "wxBitmapComboBox"},
  {wxBoolProperty, "wxBoolProperty"},
  {wxBoxSizer, "wxBoxSizer"},
  {wxBrush, "wxBrush"},
  {wxButton, "wxButton"},
  {wxCalendarCtrl, "wxCalendarCtrl"},
  {wxCaret, "wxCaret"},
  {wxCheckBox, "wxCheckBox"},
  {wxChoice, "wxChoice"},
  {wxChoicebook, "wxChoicebook"},
  {wxClientDC, "wxClientDC"},
  {wxCollapsiblePane, "wxCollapsiblePane"},
--  {wxCollapsiblePaneEvent, "wxCollapsiblePaneEvent"},
  {wxColour, "wxColour"},
--  {wxColourData, "wxColourData"},
  {wxColourDialog, "wxColourDialog"},
  {wxColourPickerCtrl, "wxColourPickerCtrl"},
--  {wxColourPickerEvent, "wxColourPickerEvent"},
  {wxColourProperty, "wxColourProperty"},
  {wxComboBox, "wxComboBox"},
  {wxConfig, "wxConfig"},
  {wxCSConv, "wxCSConv"},
  {wxCursor, "wxCursor"},
  {wxCursorProperty, "wxCursorProperty"},
--  {wxDataFormat, "wxDataFormat"},
--  {wxDateEvent, "wxDateEvent"},
  {wxDatePickerCtrl, "wxDatePickerCtrl"},
  {wxDateProperty, "wxDateProperty"},
--  {wxDateTime, "wxDateTime"},
--  {wxDC, "wxDC"},
  {wxDefButton, "wxDefButton"},
  {wxDialog, "wxDialog"},
  {wxDir, "wxDir"},
  {wxDirCtrl, "wxDirCtrl"},
  {wxDirDialog, "wxDirDialog"},
  {wxDirPickerCtrl, "wxDirPickerCtrl"},
  {wxDirProperty, "wxDirProperty"},
  {wxDirTraverser, "wxDirTraverser"},
  {wxDisplay, "wxDisplay"},
  {wxEditEnumProperty, "wxEditEnumProperty"},
  {wxEnumProperty, "wxEnumProperty"},
  {wxFileDialog, "wxFileDialog"},
  {wxFileName, "wxFileName"},
  {wxFilePickerCtrl, "wxFilePickerCtrl"},
  {wxFileProperty, "wxFileProperty"},
  {wxFlagsProperty, "wxFlagsProperty"},
  {wxFlexGridSizer, "wxFlexGridSizer"},
  {wxFloatProperty, "wxFloatProperty"},
  {wxFont, "wxFont"},
  {wxFontData, "wxFontData"},
--  {wxFontDialog, "wxFontDialog"},
--  {wxFontEnumerator, "wxFontEnumerator"},
  {wxFontPickerCtrl, "wxFontPickerCtrl"},
--  {wxFontPickerEvent, "wxFontPickerEvent"},
  {wxFontProperty, "wxFontProperty"},
  {wxFrame, "wxFrame"},
  {wxFTP, "wxFTP"},
  {wxGauge, "wxGauge"},
  {wxGraphicsBrush, "wxGraphicsBrush"},
  {wxGraphicsContext, "wxGraphicsContext"},
  {wxGraphicsMatrix, "wxGraphicsMatrix"},
  {wxGraphicsPath, "wxGraphicsPath"},
  {wxGraphicsPen, "wxGraphicsPen"},
  {wxGrid, "wxGrid"},
  {wxGridCellAttr, "wxGridCellAttr"},
  {wxGridCellAutoWrapStringEditor, "wxGridCellAutoWrapStringEditor"},
--  {wxGridCellEditor, "wxGridCellEditor"},
--  {wxGridSizer, "wxGridSizer"},
  {wxHtmlEasyPrinting, "wxHtmlEasyPrinting"},
--  {wxHtmlLinkEvent, "wxHtmlLinkEvent"},
--  {wxHtmlLinkInfo, "wxHtmlLinkInfo"},
  {wxHtmlWindow, "wxHtmlWindow"},
  {wxHTTP, "wxHTTP"},
  {wxIcon, "wxIcon"},
  {wxImage, "wxImage"},
  {wxImageFileProperty, "wxImageFileProperty"},
  {wxImageList, "wxImageList"},
--  {wxInputStream, "wxInputStream"},
  {wxIntProperty, "wxIntProperty"},
  {wxIPV4address, "wxIPV4address"},
  {wxListbook, "wxListbook"},
  {wxListBox, "wxListBox"},
  {wxListCtrl, "wxListCtrl"},
  {wxLocale, "wxLocale"},
  {wxLongStringProperty, "wxLongStringProperty"},
  {wxMask, "wxMask"},
  {wxMDIChildFrame, "wxMDIChildFrame"},
  {wxMDIClientWindow, "wxMDIClientWindow"},
  {wxMDIParentFrame, "wxMDIParentFrame"},
  {wxMediaCtrl, "wxMediaCtrl"},
  {wxMemoryDC, "wxMemoryDC"},
  {wxMenu, "wxMenu"},
  {wxMenuBar, "wxMenuBar"},
  {wxMenuItem, "wxMenuItem"},
--  {wxMouseEvent, "wxMouseEvent"},
  {wxMultiChoiceDialog, "wxMultiChoiceDialog"},
  {wxMultiChoiceProperty, "wxMultiChoiceProperty"},
  {wxNotebook, "wxNotebook"},
--  {wxOutputStream, "wxOutputStream"},
  {wxPageSetupDialog, "wxPageSetupDialog"},
--  {wxPageSetupDialogData, "wxPageSetupDialogData"},
  {wxPaintDC, "wxPaintDC"},
  {wxPanel, "wxPanel"},
  {wxPasswordEntryDialog, "wxPasswordEntryDialog"},
  {wxPen, "wxPen"},
--  {wxPlatformInfo, "wxPlatformInfo"},
  {wxPopupTransientWindow, "wxPopupTransientWindow"},
  {wxPopupWindow, "wxPopupWindow"},
  {wxPrintDialog, "wxPrintDialog"},
  {wxPrinter, "wxPrinter"},
  {wxPrintout, "wxPrintout"},
  {wxPrintPreview, "wxPrintPreview"},
  {wxProcess, "wxProcess"},
--  {wxProcessEvent, "wxProcessEvent"},
  {wxProgressDialog, "wxProgressDialog"},
  {wxPropertyCategory, "wxPropertyCategory"},
  {wxPropertyGrid, "wxPropertyGrid"},
  {wxRadioBox, "wxRadioBox"},
  {wxRadioButton, "wxRadioButton"},
  {wxRegEx, "wxRegEx"},
  {wxRichTextAttr, "wxRichTextAttr"},
  {wxRichTextCtrl, "wxRichTextCtrl"},
--  {wxRichTextEvent, "wxRichTextEvent"},
  {wxRichTextFormattingDialog, "wxRichTextFormattingDialog"},
  {wxRichTextPrinting, "wxRichTextPrinting"},
  {wxScrollBar, "wxScrollBar"},
  {wxScrolledWindow, "wxScrolledWindow"},
  {wxSingleChoiceDialog, "wxSingleChoiceDialog"},
  {wxSlider, "wxSlider"},
  {wxSocketClient, "wxSocketClient"},
  {wxSocketServer, "wxSocketServer"},
  {wxSound, "wxSound"},
  {wxSpinCtrl, "wxSpinCtrl"},
  {wxSplashScreen, "wxSplashScreen"},
--  {wxSplitterEvent, "wxSplitterEvent"},
  {wxSplitterWindow, "wxSplitterWindow"},
  {wxStaticBitmap, "wxStaticBitmap"},
  {wxStaticBox, "wxStaticBox"},
  {wxStaticBoxSizer, "wxStaticBoxSizer"},
  {wxStaticLine, "wxStaticLine"},
  {wxStaticText, "wxStaticText"},
  {wxStatusBar, "wxStatusBar"},
  {wxStdDialogButtonSizer, "wxStdDialogButtonSizer"},
  {wxStringProperty, "wxStringProperty"},
  {wxSystemColourProperty, "wxSystemColourProperty"},
  {wxTaskBarIcon, "wxTaskBarIcon"},
  {wxTextAttr, "wxTextAttr"},
  {wxTextAttrEx, "wxTextAttrEx"},
  {wxTextCtrl, "wxTextCtrl"},
--  {wxTextDataObject, "wxTextDataObject"},
  {wxTextEntryDialog, "wxTextEntryDialog"},
  {wxTimer, "wxTimer"},
  {wxToggleButton, "wxToggleButton"},
  {wxToolBar, "wxToolBar"},
  {wxToolBarItem, "wxToolBarItem"},
  {wxToolbook, "wxToolbook"},
  {wxTreebook, "wxTreebook"},
  {wxTreeCtrl, "wxTreeCtrl"},
--  {wxTreeEvent, "wxTreeEvent"},
  {wxUIntProperty, "wxUIntProperty"},
--  {wxUpdateUIEvent, "wxUpdateUIEvent"},
  {wxURL, "wxURL"},
  {wxWindow, "wxWindow"},
  {wxWizard, "wxWizard"},
  {wxWizardPage, "wxWizardPage"},
  {wxXmlDocument, "wxXmlDocument"},
  {wxXmlNode, "wxXmlNode"},
  {wxXmlProperty, "wxXmlProperty"},
  {wxXmlResource, "wxXmlResource"}
}

public sequence wxeuDir = "wxeu"&SLASH
sequence resourceDir = wxeuDir&"res"&SLASH
sequence toolButtons = {}

------------------------------------------------------------------------------

-- get the ID of a widget
export function getID(integer id)
  return registry[id][REG_ID]
end function

------------------------------------------------------------------------------

-- get the class of a widget
export function getClass(integer id)
  return registry[id][REG_CLASS]
end function

------------------------------------------------------------------------------

-- get the children of a widget
export function getChildren(integer id)
  return registry[id][REG_CHILDREN]
end function

------------------------------------------------------------------------------

-- get the parameters of a widget
export function getParams(integer id)
  return registry[id][REG_PARAMS]
end function

------------------------------------------------------------------------------

-- set the children of a widget
export procedure setChildren(integer id, sequence children)
  registry[id][REG_CHILDREN] = children
end procedure

------------------------------------------------------------------------------

-- set the parameters of a widget
export procedure setParams(integer id, sequence parameters)
  registry[id][REG_PARAMS] = parameters
end procedure

------------------------------------------------------------------------------

-- find a Widget with its handle or the handle of its child
export function findWidget(integer id)
  if id = 0 then return 0 end if
  sequence s = find_nested(id, registry)
--  analyzeObject(s, "s", f_debug)
  if length(s) then
    return s[1]
  else
    return 0
  end if
end function

------------------------------------------------------------------------------

-- get the class name of a widget
export function getClassName(integer id)
  object class = getClass(id)
  return vlookup(class, WX_EUPHORIA_CLASSES, 1, 2, "Unknown")
end function

------------------------------------------------------------------------------

-- create a widget (visual component) and store it in registry
export function createWidget( integer class, integer parent, sequence text="", integer x=0, integer y=0,
                        integer w=0, integer h=0, object params={}, object tip=TIP_NONE )
  atom id
  sequence children = {}, parms = {}

  if atom(tip) and (tip=TIP_AUTO) then
    tip = vlookup(class, WX_EUPHORIA_CLASSES, 1, 2, "Unknown")
  end if

  switch class do
  case wxButton, wxBitmapButton, wxDefButton, wxToggleButton then
    id = wx:create( class, {getID(parent), -1, text, x, y, w, h} )
    if atom(params) then enable(id, params) end if
  case wxChoice then
    id = wx:create( class, {getID(parent), -1, x, y, w, h, params} )
  case wxComboBox then
    id = wx:create( class, {getID(parent), -1, text, x, y, w, h, params[1], wxCB_DROPDOWN} )
    set_selection(id, params[2]-1)
  case wxFrame then
    if parent then
      id = wx:create( class, {getID(parent), -1, text, x, y, w, h} )
    else  
      id = wx:create( class, {0, -1, text, x, y, w, h} )
    end if
  case wxGrid then
    id = wx:create( class, {getID(parent), -1, x, y, w, h, 0, 0, 0} )
    -- change the color of the grid lines to black
    set_grid_line_color( id, wxBLACK )
  case wxListBox then
    id = wx:create( class, {getID(parent), -1, x, y, w, h} )
    add_item(id, params)
  case wxMenu then
    id = wx:create( class, {getID(parent), text} )
  case wxMenuBar then
    id = wx:create( class, {getID(parent)} )
  case wxMenuItem then
    id = wx:create( class, {getID(parent), params, text} )
  case wxPanel then
    if sequence(params) and length(params) then
      id = wx:create( class, {getID(parent), -1, -1, -1, -1, -1, sum(params)} )
    else
      id = wx:create( class, {getID(parent)} )
    end if
  case wxStaticBox then
    id = wx:create( class, {getID(parent), -1, text, x, y, w, h, params} )
  case wxStatusBar then
    if length(text) and sequence(text[1]) then
      id = create( wxStatusBar, {getID(parent), length(text) })
      for i = 1 to length(text) do
        set_status_text(id, text[i], i-1)
      end for
    else  
      id = create( wxStatusBar, {getID(parent), 1 })
      set_status_text(id, text, 0)
    end if
    set_status_fields(id, params )
    set_status_styles(id, {wxSB_NORMAL, wxSB_FLAT})
  case wxStaticText, wxTextCtrl then
    if sequence(params) and length(params) then
      id = wx:create( class, {getID(parent), -1, text, x, y, w, h, sum(params)} )
    else
      id = wx:create( class, {getID(parent), -1, text, x, y, w, h} )
    end if
  case wxToolBar then
    id = wx:create( class, {getID(parent), -1, x, y, w, h, sum(params)} )
  case else
    id = wx:create( class, {getID(parent), -1, text, x, y, w, h} )
  end switch
  if find(class, {wxAuiNotebook, wxBitmapButton, wxBitmapComboBox,
      wxButton, wxCalendarCtrl, wxCheckBox, wxChoice, wxChoicebook,
      wxCollapsiblePane, wxColourPickerCtrl, wxComboBox,
      wxDatePickerCtrl, wxDefButton, wxDialog, wxDirCtrl,
      wxDirPickerCtrl, wxFilePickerCtrl, wxFontPickerCtrl, wxFrame,
      wxGauge, wxGrid, wxHtmlWindow, wxListbook, wxListBox,
      wxListCtrl, wxMDIChildFrame, wxMDIClientWindow, wxMDIParentFrame,
      wxMediaCtrl, wxMenu, wxMenuBar, wxMenuItem, wxNotebook, wxPanel,
      wxPopupWindow, wxRadioBox, wxRadioButton, wxRichTextCtrl,
      wxScrollBar, wxScrolledWindow, wxSlider, wxSpinCtrl,
      wxSplashScreen, wxSplitterWindow, wxStaticBitmap, wxStaticBox,
      wxStaticLine, wxStaticText, wxStatusBar, wxTaskBarIcon,
      wxTextCtrl, wxToggleButton, wxToolBar, wxToolBarItem,
      wxToolbook, wxTreebook, wxTreeCtrl, wxWindow, wxWizard,
      wxWizardPage}) then
    if sequence(tip) then set_tip(id, tip) end if
  end if

  registry = append(registry, {id, class, children, parms, tip})
  return length(registry)
end function

------------------------------------------------------------------------------

-- create a non-visual component or a dialog and store it in registry
export function createEx( integer class, object params={} )
  switch class do
  case wxFont then
  case wxTimer, wxFileDialog, wxDirDialog then
    params[1] = getID(params[1])
  end switch
  
  atom id = wx:create( class, params)
  registry = append(registry, {id, class, params})
  return length(registry)
end function

------------------------------------------------------------------------------

-- set the event handler of a widget
export procedure setHandler(integer widget, integer event, integer event_routine=0, integer id=-1)
  atom class = getClass(widget)
  atom this = getID(widget)
  switch class do
  case wxFrame then
    switch event do
    case wxEVT_CLOSE_WINDOW then
      set_event_handler(this, -1, event, event_routine)
    case wxEVT_SHOW then
      set_event_handler(this, get_id(this), event, event_routine)
    case wxEVT_TIMER then
      set_event_handler(this, id, event, event_routine)
    end switch
  case wxButton, wxBitmapButton, wxDefButton, wxToggleButton then
    set_event_handler(this, get_id(this), event, event_routine)
  case wxMenu then
    set_event_handler(this, id, event, event_routine)
  case else
    set_event_handler(this, -1, event, event_routine)
  end switch
end procedure

------------------------------------------------------------------------------

-- set the text of a widget
export procedure setText(integer widget, sequence text, object item=0)
  atom class = getClass(widget)
  atom id = getID(widget)
  switch class do
    case wxAuiNotebook then
      aui_notebook_set_page_text( id, item-1, text )
    case wxButton, wxBitmapButton, wxDefButton, wxToggleButton then
      set_button_label( id, text )
    case wxChoice, wxComboBox, wxBitmapComboBox then
      set_string( id, item-1, text )
    case wxChoicebook then
      set_page_text( id, item-1, text )
    case wxFrame then
      set_title( id, text )
    case wxGrid, wxGridCellAttr, wxGridCellAutoWrapStringEditor /*, wxGridCellEditor */ then
      if atom(item) then return end if
      switch item[1] do
      case wxROW then
        set_row_label( id, item[2]-1, text )
      case wxCOL then
        set_col_label( id, item[2]-1, text )
      case wxCELL then
        set_cell_value(id, text, item[2]-1, item[3]-1)
      end switch
    case wxListbook then
      set_page_text( id, item, text )
    case wxListBox, wxListCtrl then
      set_string( id, item-1, text )  -- une des deux est bonne
    case wxMenu, wxMenuBar, wxMenuItem then
      set_menu_label( id, item-1, text )
    case wxNotebook then
      set_page_text( id, item-1, text )
    case wxPanel then
      set_title( id, text )
    case wxRadioBox then
      set_label( id, text )
    case wxRichTextCtrl then
      set_text( id, text )
    case wxRichTextPrinting then
      switch item[1] do
      case wxFOOTER then
        set_richprint_footer_text( id, text, item[2]-1, item[3]-1 )
      case wxHEADER then
        set_richprint_header_text( id, text, item[2]-1, item[3]-1 )
      end switch
    case wxScrolledWindow then
      set_title( id, text )
    case wxStaticText then
      set_label( id, text )
    case wxStatusBar then
      if item then
        set_status_text( id, text, item-1)
      else
        set_status_text( id, text, 0 )
      end if
    case wxTextAttrEx then
      set_textattrex_bullet_text( id, text )
    case wxTextCtrl then
      set_text( id, text )
    case wxToolBar, wxToolBarItem then
      set_tool_label( id, item-1, text )   
    case wxToolbook, wxTreebook then
      set_page_text( id, item-1, text )
    case wxTreeCtrl then
      set_tree_item_text( id, item-1, text )   
    case wxWindow then
      set_title( id, text )
  end switch
end procedure

------------------------------------------------------------------------------

-- get the text of a widget
export function getText(integer widget, object item=0)
  atom class = getClass(widget)
  atom id = getID(widget)
  switch class do
    case wxButton, wxBitmapButton, wxDefButton, wxToggleButton then
      return get_label( id )
    case wxChoice, wxComboBox, wxBitmapComboBox then
      return get_combo_string( id )
    case wxFrame then
      return get_label( id )
    case wxGrid, wxGridCellAttr, wxGridCellAutoWrapStringEditor /*, wxGridCellEditor */ then
      if atom(item) then return "" end if
      switch item[1] do
      case wxCOL then
        return get_col_label( id, item[2]-1 )
      case wxCELL then
        return get_cell_value(id, item[2]-1, item[3]-1)
      end switch
    case wxListBox, wxListCtrl then
      return get_string_selection( id )
    case wxMenu, wxMenuBar, wxMenuItem then
      return get_label( id )
    case wxPanel then
      return get_label( id )
    case wxRadioBox then
      return get_label( id )
    case wxRichTextCtrl then
      return get_text_value( id )
    case wxRichTextPrinting then
      switch item do
      case wxFOOTER then
        return get_richprint_footer_text( id )
      case wxHEADER then
        return get_richprint_header_text( id )
      end switch
    case wxScrolledWindow then
      return get_label( id )
    case wxStaticText then
      return get_label( id )
    case wxStatusBar then
      if item then
        return get_status_text( id, item-1)
      else
        return get_status_text( id, 0 )
      end if
    case wxTextAttrEx then
      return get_textattrex_bullet_text( id )
    case wxTextCtrl then
      return get_text_value(id)
    case wxToolBarItem then
      return get_label( id )   
    case wxTreeCtrl then
      return get_tree_selected_text( id )   
    case wxWindow then
      return get_label( id )
  end switch
  return 0
end function

------------------------------------------------------------------------------

-- get the number of text lines of a widget
export function getLinesCount(integer widget)
  atom class = getClass(widget)
  atom id = getID(widget)
  switch class do
    case wxTextCtrl then
      return get_text_lines(id)
  end switch
  return 0
end function

------------------------------------------------------------------------------

-- select text in a widget
export procedure selectText(integer widget, integer fromPos, integer toPos)
  atom id = getID(widget)
  set_text_selection(id, fromPos-1, toPos-1)
end procedure

------------------------------------------------------------------------------

-- scrolls text until position shows
export procedure showText(integer widget, integer fromPos)
  atom id = getID(widget)
  show_text_position (id, fromPos-1)
end procedure

------------------------------------------------------------------------------

-- append text to a widget
export procedure appendText(integer widget, sequence text)
  atom id = getID(widget)
  append_text(id, text)
end procedure

------------------------------------------------------------------------------

-- returns the insertion point of a text control
export function getInsertionPoint(integer widget)
  atom id = getID(widget)
  return get_text_insertion_point(id)
end function

------------------------------------------------------------------------------

-- displays a dialog
export function messageBox(sequence text, sequence title, atom ok)
  return message_box(text, title, ok)
end function

------------------------------------------------------------------------------

-- returns a new ID
export function newID()
  return new_id()
end function

------------------------------------------------------------------------------

-- shows a modal dialog
export function showModal(integer widget)
  atom id = getID(widget)
  return show_modal(id)
end function

------------------------------------------------------------------------------

-- set the default directory of an wxFileDialog or wxDirDialog
export procedure setDirectory(integer widget, sequence path)
  atom id = getID(widget)
  set_fd_directory ( id, path )
end procedure

------------------------------------------------------------------------------

-- returns the path of the selected wxDirDialog
export function getDirectory(integer widget)
  atom id = getID(widget)
  return get_dir_path(id)
end function

------------------------------------------------------------------------------

-- returns the full pathname of the selected wxFileDialog
export function getFullPathName(integer widget)
  atom id = getID(widget)
  return get_fd_path(id)
end function

------------------------------------------------------------------------------

-- returns the (short) filename of the selected wxFileDialog
export function getFileName(integer widget)
  atom id = getID(widget)
  return get_fd_filename(id)
end function

------------------------------------------------------------------------------

-- closes application
export procedure closeApp()
  exit_main()
end procedure

------------------------------------------------------------------------------

-- refresh widget
export procedure refreshWidget(integer widget)
  atom id = getID(widget)
  refresh_window(id)
end procedure

------------------------------------------------------------------------------

-- process wxEuphoria events
export function doEvents(integer onlyIfNeeded)
  return app_yield (onlyIfNeeded)
end function

------------------------------------------------------------------------------

-- apply a sizer to a widget
export procedure setSizer(integer widget, integer sizer)
  atom id = getID(widget)
  atom fs = getID(sizer)
  analyzeObject(id, "id", f_debug)
  analyzeObject(fs, "fs", f_debug)
  set_sizer(id, fs)
end procedure

------------------------------------------------------------------------------

-- add an item (window or sizer) to a sizer
export procedure addToSizer(integer widget, integer item, atom option, atom flag, atom border)
  atom class = getClass(item)
  atom id = getID(widget)
  atom sz = getID(item)
  switch class do
    case wxBoxSizer, wxFlexGridSizer, /* wxGridSizer,*/ wxStaticBoxSizer then
      add_sizer_to_sizer ( id, sz, option, flag, border )
    case else
      add_window_to_sizer ( id, sz, option, flag, border )
  end switch
end procedure


------------------------------------------------------------------------------

-- set the enabled state of a widget
export procedure setEnable(integer widget, integer state, integer item=0)
  atom class = getClass(widget)
  atom id = getID(widget)
  if class =  wxToolBar then
    enable_tool( id, item, state )
  elsif find(class, {wxAuiNotebook, wxBitmapButton, wxBitmapComboBox, wxButton,
                     wxCalendarCtrl, wxCheckBox, wxChoice, wxChoicebook,
                     wxCollapsiblePane, wxColourPickerCtrl, wxComboBox,
                     wxDatePickerCtrl, wxDialog, wxDirCtrl, wxDirPickerCtrl,
                     wxFilePickerCtrl, wxFontPickerCtrl, wxFrame, wxGauge, wxGrid,
                     wxHtmlWindow, wxListbook, wxListBox, wxListCtrl, wxMDIChildFrame,
                     wxMDIClientWindow, wxMDIParentFrame, wxMediaCtrl, wxMenu,
                     wxMenuBar, wxMenuItem, wxNotebook, wxPanel, wxPopupWindow,
                     wxRadioBox, wxRadioButton, wxRichTextCtrl, wxScrollBar,
                     wxScrolledWindow, wxSlider, wxSpinCtrl, wxSplashScreen,
                     wxSplitterWindow, wxStaticBitmap, wxStaticBox, wxStaticLine,
                     wxStaticText, wxStatusBar, wxTaskBarIcon, wxTextCtrl,
                     wxToggleButton, wxToolBar, wxToolBarItem, wxToolbook, wxTreebook,
                     wxTreeCtrl, wxWindow, wxWizard, wxWizardPage}) then
    enable(id, state)
  end if
end procedure

------------------------------------------------------------------------------

-- set the visible state of a widget
export procedure showWidget(integer widget, integer state=0)
  atom class = getClass(widget)
  atom id = getID(widget)
  switch class do
    case wxFrame, wxMDIParentFrame then
      show_window(id, state)
    case wxToolBar then
      show_toolbar( id )
  end switch
end procedure

------------------------------------------------------------------------------

-- add a separator to a widget
export procedure addSeparator(integer widget, integer pos=0)
  atom class = getClass(widget)
  atom id = getID(widget)
  switch class do
    case wxMenu then
      if pos = 0 then
        append_separator(id)
      else
        insert_separator(id, pos-1 )
      end if
    case wxToolBar then
      add_separator(id)
  end switch
end procedure

------------------------------------------------------------------------------

-- start a timer
export procedure startTimer(integer ID, atom msec, integer oneshot)
  atom id = getID(ID)
  start_timer(id, msec, oneshot)
end procedure

------------------------------------------------------------------------------

-- stop a timer
export procedure stopTimer(integer ID)
  atom id = getID(ID)
  stop_timer(id)
end procedure

------------------------------------------------------------------------------

-- set the focus to a widget
export procedure setFocus(integer widget)
  atom id = getID(widget)
  set_focus(id)
end procedure

------------------------------------------------------------------------------

-- select an item of a widget
export procedure setSelection(integer widget, integer item)
  atom id = getID(widget)
  set_selection(id, item)
end procedure

------------------------------------------------------------------------------

-- set the font of a widget
export procedure setFont(integer ID, integer font)
  atom id = getID(ID)
  atom fnt = getID(font)
  set_default_font(id, fnt)
end procedure

-----------------------------------------------------------------------------

-- add a button to a toolbar
export function addToolButton(integer widget, sequence text, object xpm, sequence hint)
  atom img = create( wxBitmap, {BM_XPM, xpm} )
  atom id = new_id()
  sequence children = getChildren(widget)
  toolButtons = append(toolButtons, add_tool( getID(widget), id, img, img, text, text, hint, wxITEM_NORMAL ))
  children = append(children, id)
  setChildren(widget, children)
  enable_tool(getID(widget), id, wxTrue)
  return id
end function

------------------------------------------------------------------------------

-- run event loop
export procedure winMain(integer widget)
  atom id = getID(widget)
  wxMain( id )
end procedure

------------------------------------------------------------------------------

-- set up grid columns
export procedure gridSetup(integer widget, sequence fields)
  atom grid = getID(widget)
  integer lg = length(fields)
  set_grid_size ( grid, 1, lg )
  for i = 1 to lg  do
    set_col_label( grid, i-1, fields[i][FIELD_NAME] )
    set_col_size( grid, i-1, fields[i][FIELD_SIZE] )
    set_col_alignment( grid, i-1, fields[i][FIELD_ALIGNMENT][1], fields[i][FIELD_ALIGNMENT][2] )
    switch fields[i][FIELD_TYPE] do
      case wxBOOL then set_col_format_bool( grid, i-1 )
      case wxFLOAT then set_col_format_float( grid, i-1, fields[i][FIELD_WIDTH], fields[i][FIELD_PRECISION] )
      case wxINTEGER then set_col_format_number( grid, i-1 )
    end switch
  end for
  force_refresh_grid ( grid )
end procedure

------------------------------------------------------------------------------

-- clears a grid preserving the column headers
export procedure gridClear(integer widget)
  atom grid = getID(widget)
  integer lg = get_number_cols ( grid )
  sequence colLabel = {}
  for i = 1 to lg  do
    colLabel = append( colLabel, get_col_label(grid, i-1) )
  end for
  clear_grid ( grid )
  set_grid_size ( grid, 1, lg )
  for i = 1 to lg  do
    set_col_label( grid, i-1, colLabel[i] )
  end for
end procedure

------------------------------------------------------------------------------

-- load data from a CSV file to a grid
export procedure gridLoad(atom widget, sequence filename)
  atom grid = getID(widget)
  sequence s = readCSV(filename)
--  analyzeSequence(s, "s", f_debug)
  integer nbRows = length(s)
  integer nbCols = length(s[1])
  set_grid_size ( grid, nbRows+2, nbCols )
  for c = 1 to nbCols do
    set_col_label ( grid, s[1][c], c-1 )
  end for
  for r = 2 to nbRows do
    for c = 1 to nbCols do
      set_cell_value ( grid, s[r][c], r-2, c-2 )
    end for
  end for
  force_refresh_grid ( grid )
  set_focus(grid)
end procedure

------------------------------------------------------------------------------

-- saves grid data to a CSV file
export procedure gridSave(atom widget, sequence filename)
  atom grid = getID(widget)
  integer nbRows = get_number_rows ( grid )
  integer nbCols = get_number_cols ( grid )
  sequence s = {}
  sequence colLabel = {}
  for i = 1 to nbCols  do
    colLabel = append( colLabel, get_col_label(grid, i-1) )
  end for
  s = append(s, colLabel)
  for r = 1 to nbRows do
    sequence t = {}
    for c = 1 to nbCols do
      t = append(t, get_cell_value ( grid, r-1, c-1 ) )
    end for
    s = append(s, t)
  end for
--  AnalyseSequence(s, "s", f_debug)
  saveCSV(filename, s)
end procedure

------------------------------------------------------------------------------

-- adds a row with specified data
export procedure gridAddRow(integer widget, sequence data)
  atom grid = getID(widget)
  append_rows ( grid, 1 )
  integer row = get_number_rows( grid )
  begin_batch(grid)
  for i = 1 to length(data) do
    set_cell_value(grid, data[i], row-1, i-1)
  end for
  end_batch(grid)
end procedure

------------------------------------------------------------------------------

-- adds a new blank row when clicking on the last row
export procedure setEditText(atom widget)
  atom grid = getID(widget)
  if grid_pos ( grid ) = get_number_rows ( grid ) then     
    append_rows ( grid, 1 )
  end if
end procedure

------------------------------------------------------------------------------

-- set the height of the column header row
export procedure setColHeaderSize(integer widget, integer size)
  atom id = getID(widget)
  set_col_label_size( id, size )
end procedure

------------------------------------------------------------------------------

-- set the width of the row header column
export procedure setRowHeaderSize(integer widget, integer size)
  atom id = getID(widget)
  set_row_label_size( id, size )
end procedure

------------------------------------------------------------------------------

-- set the selection mode of a grid
export procedure setSelectionMode(integer widget, integer mode)
  atom id = getID(widget)
  set_grid_selection_mode( id, mode )
end procedure
