-- klrconio.e
-- text mode Euphoria routines for the Windows console based on kernel32

-- globals
-- cPuts(sequence s)
-- cPrintf(sequence format, object data)
-- gotoXY(integer x, integer y)
-- clrScr()
-- clrEol()
-- insLine()
-- delLine()
-- whereX() returns integer
-- whereY() returns integer
-- getScreenHeight() returns integer
-- getScreenWidth() returns integer
-- getText(int x1, int y1, int x2, int y2, atom savebuf)
-- putText(int x1, int y1, int x2, int y2, atom savebuf)
-- setWindowTitle(sequence s)
-- cBreakOn()  allows Ctrl_c to stop execution
-- cBreakOff()  Ctrl_c will not stop execution (but Ctrl_Break will)
-- getCh() reads but does not echo a keypress and returns a
--   sequence of five integers
--   (1)  ascii code, or zero if none
--   (2)  dos extended key code, or zero if key is ascii
--   (3)  win32 virtual key code
--   (4)  shift state, considers capslock and shift keys
--   (5)  control key state, one for set, zero for not
-- cGets() echos and returns a character sequence from the keyboard
-- seq2int(sequence s) returns object
-- seq2float(sequence s) returns object
--   the object is an atom if s contains (only) a valid number, or else
--   it is a sequence containing the characters "nan"

-- notes
-- screen coordinates start at 1
-- getText and putText save and restore (a portion of) the screen
--   example--save and restore the entire screen:
--   integer savex, savey, maxx, maxy
--   atom savebuf
--   maxx=getScreenWidth()
--   maxy=getScreenHeight()
--   savebuf=allocate(2*maxx*maxy)
--   savex=whereX()
--   savey=whereY()
--   getText(1,1,maxx,maxy,savebuf)
--   clrScr()
--   ...
--   putText(1,1,maxx,maxy,savebuf)
--   gotoXY(savex,savey)
--   free(savebuf)

include std/dll.e
include std/machine.e
include std/console.e
include std/convert.e

--constants, none of these are public

constant STD_OUTPUT_HANDLE = #FFFFFFF5
constant STD_INPUT_HANDLE = #FFFFFFF6

atom kernel32
  kernel32 = open_dll("kernel32.dll")

constant GetStdHandle =
  define_c_func(kernel32,"GetStdHandle",
  {C_UINT},C_INT)

constant SetConsoleTitle =
  define_c_proc(kernel32,"SetConsoleTitleA",
  {C_POINTER})

constant WriteConsole =
  define_c_proc(kernel32,"WriteConsoleA",
  {C_INT,C_POINTER,C_INT,C_POINTER,C_POINTER})

constant GetConsoleScreenBufferInfo =
  define_c_proc(kernel32,"GetConsoleScreenBufferInfo",
  {C_INT,C_POINTER})

constant FillConsoleOutputAttribute =
  define_c_proc(kernel32,"FillConsoleOutputAttribute",
  {C_INT,C_CHAR,C_INT,C_UINT,C_POINTER})

constant FillConsoleOutputCharacter =
  define_c_proc(kernel32,"FillConsoleOutputCharacterA",
  {C_INT,C_CHAR,C_INT,C_UINT,C_POINTER})

constant SetConsoleCursorPosition =
  define_c_proc(kernel32,"SetConsoleCursorPosition",
  {C_INT,C_UINT})

constant ScrollConsoleScreenBuffer =
  define_c_proc(kernel32,"ScrollConsoleScreenBufferA",
  {C_INT,C_POINTER,C_POINTER,C_UINT,C_POINTER})

constant ReadConsoleOutput =
  define_c_proc(kernel32,"ReadConsoleOutputA",
  {C_INT,C_POINTER,C_UINT,C_UINT,C_POINTER})

constant WriteConsoleOutput =
  define_c_proc(kernel32,"WriteConsoleOutputA",
  {C_INT,C_POINTER,C_UINT,C_UINT,C_POINTER})

constant ReadConsoleInput =
  define_c_proc(kernel32,"ReadConsoleInputA",
  {C_INT,C_POINTER,C_INT,C_POINTER})

constant
  SBI_BYTES = 22,
  SBI_SIZEX = 0,      -- size of the console screen buffer, in character columns
  SBI_SIZEY = 2,      -- size of the console screen buffer, in character rows
  SBI_CPOSX = 4,      -- column coordinates of the cursor in the console screen buffer
  SBI_CPOSY = 6,      -- row coordinates of the cursor in the console screen buffer
  SBI_ATTR = 8,       -- attributes of the characters written to a screen buffer
  SBI_WINX1 = 10,     -- console screen buffer upper-left column
  SBI_WINX2 = 12,     -- console screen buffer upper-left row
  SBI_WINY1 = 14,     -- console screen buffer lower-right column
  SBI_WINY2 = 16,     -- console screen buffer lower-right row
  SBI_MAX_X = 18,     -- maximum size of the console window, in character columns
  SBI_MAX_Y = 20      -- maximum size of the console window, in character rows

public constant
  BUFFER_COLS = 1,
  BUFFER_ROWS = 2,
  CURSOR_COL = 3,
  CURSOR_ROW = 4,
  CHAR_ATTRS = 5,
  WIN_LEFT = 6,
  WIN_RIGHT = 7,
  WIN_TOP = 8,
  WIN_BOTTOM = 9,  
  MAX_COLS  = 10,
  MAX_ROWS = 11

constant
  INREC_BYTES  = 20,
  INREC_EVENT  = 0,
  INREC_STRUCT = 4

public constant
  -- input event
  KEY_EVENT                = #01,  -- keyboard event, a KEY_EVENT_RECORD structure is returned.
  MOUSE_EVENT              = #02,  -- mouse movement or button press event, a MOUSE_EVENT_RECORD structure is returned.
  WINDOW_BUFFER_SIZE_EVENT = #04,  -- window resized, a WINDOW_BUFFER_SIZE_RECORD structure is returned (coords).
  MENU_EVENT               = #08,  -- internal usage, a MENU_EVENT_RECORD structure is returned (uint: command ID). Should be ignored.
  FOCUS_EVENT              = #10  -- internal usage, a FOCUS_EVENT_RECORD structure is returned (boolean). Should be ignored.

constant
  -- key event structure
  KEY_DOWN          = 0,
  REPEAT_COUNT      = 4,
  VIRTUAL_KEYCODE   = 6,
  VIRTUAL_SCANCODE  = 8,
  KEY_CHAR          = 10,
  CONTROL_KEY_STATE = 12

constant
  -- mouse event structure
  MOUSE_POSITION_X  = 0,
  MOUSE_POSITION_Y  = 2,
  BUTTON_STATE      = 4,
  MOUSE_CONTROL_KEYS= 8,
  EVENT_FLAGS       = 12

constant
  -- window buffer size event structure
  WINDOW_SIZE_X     = 0,
  WINDOW_SIZE_Y     = 2

constant
  -- control keys state
  RIGHT_ALT_PRESSED      = #0001,  -- right ALT key pressed
  LEFT_ALT_PRESSED       = #0002,  -- left ALT key pressed
  RIGHT_CTRL_PRESSED     = #0004,  -- right CTRL key pressed
  LEFT_CTRL_PRESSED      = #0008,  -- left CTRL key pressed
  SHIFT_PRESSED          = #0010,  -- SHIFT key pressed
  NUMLOCK_ON             = #0020,  -- NUM LOCK light is on
  SCROLLLOCK_ON          = #0040,  -- SCROLL LOCK light is on
  CAPSLOCK_ON            = #0080,  -- CAPS LOCK light is on
  ENHANCED_KEY           = #0100   -- key is enhanced

constant
  -- mouse button state
  FROM_LEFT_1ST_BUTTON_PRESSED = #0001,  -- leftmost mouse button.
  RIGHTMOST_BUTTON_PRESSED     = #0002,  -- rightmost mouse button.
  FROM_LEFT_2ND_BUTTON_PRESSED = #0004,  -- second button from the left.
  FROM_LEFT_3RD_BUTTON_PRESSED = #0008,  -- third button from the left.
  FROM_LEFT_4TH_BUTTON_PRESSED = #0010   -- fourth button from the left.

constant
  -- mouse event flags
  MOUSE_MOVED    = #0001,  -- A change in mouse position occurred.
  DOUBLE_CLICK   = #0002,  -- second click (button press) of a double-click occurred.
                           -- first click is returned as a regular button-press event.
  MOUSE_WHEELED  = #0004,  -- vertical mouse wheel was moved.
                           -- If the high word of the dwButtonState member contains a positive value,
                           -- the wheel was rotated forward, away from the user.
                           -- Otherwise, the wheel was rotated backward, toward the user.
  MOUSE_HWHEELED = #0008   -- horizontal mouse wheel was moved.
                           -- If the high word of the dwButtonState member contains a positive value,
                           -- the wheel was rotated to the right.
                           -- Otherwise, the wheel was rotated to the left.

constant
  VK_SHIFT    = #10,
  VK_CONTROL  = #11,
  VK_MENU     = #12,
  VK_CAPITAL  = #14,
  VK_PRIOR    = #21,
  VK_NEXT     = #22,
  VK_END      = #23,
  VK_HOME     = #24,
  VK_LEFT     = #25,
  VK_UP       = #26,
  VK_RIGHT    = #27,
  VK_DOWN     = #28,
  VK_SNAPSHOT = #2C,
  VK_INSERT   = #2D,
  VK_DELETE   = #2E,
  VK_F1       = #70,
  VK_F2       = #71,
  VK_F3       = #72,
  VK_F4       = #73,
  VK_F5       = #74,
  VK_F6       = #75,
  VK_F7       = #76,
  VK_F8       = #77,
  VK_F9       = #78,
  VK_F10      = #79,
  VK_F11      = #7A,
  VK_F12      = #7B,
  VK_NUMLOCK  = #90,
  VK_SCROLL   = #91


-- functions and procedures (not all of these are public)

function coord(integer x, integer y)
  return (x+y*#10000)
end function

public procedure cPuts(sequence s)
  integer hout=c_func(GetStdHandle,{STD_OUTPUT_HANDLE})
  atom cs=allocate_string(s)
  atom pdm=allocate(4)
  c_proc(WriteConsole,{hout,cs,length(s),pdm,NULL})
  free(pdm)
  free(cs)
end procedure

public procedure cPrintf(sequence fs, object data)
  cPuts(sprintf(fs, data))
end procedure

public function getConsoleInfo()
  integer hout=c_func(GetStdHandle,{STD_OUTPUT_HANDLE})
  atom sbi=allocate(SBI_BYTES)
  c_proc(GetConsoleScreenBufferInfo,{hout,sbi})
  integer nx =   peek(sbi+SBI_SIZEX)
  integer ny =   peek(sbi+SBI_SIZEY)
  integer cx = 1+peek(sbi+SBI_CPOSX)
  integer cy = 1+peek(sbi+SBI_CPOSY)
  integer attr = peek(sbi+SBI_ATTR)
  integer x1 = 1+peek(sbi+SBI_WINX1)
  integer x2 = 1+peek(sbi+SBI_WINX2)
  integer y1 = 1+peek(sbi+SBI_WINY1)
  integer y2 = 1+peek(sbi+SBI_WINY2)
  integer maxx = peek(sbi+SBI_MAX_X)
  integer maxy = peek(sbi+SBI_MAX_Y)
  free(sbi)
  return {nx,ny,cx,cy,attr,x1,x2,y1,y2,maxx,maxy}
end function

public procedure clrScr()
  sequence ci = getConsoleInfo()
  integer ntot=ci[BUFFER_COLS]*ci[BUFFER_ROWS]
  integer hout=c_func(GetStdHandle,{STD_OUTPUT_HANDLE})
  atom ncw=allocate(4)
  c_proc(FillConsoleOutputAttribute,{hout,ci[CHAR_ATTRS],ntot,coord(0,0),ncw})
  c_proc(FillConsoleOutputCharacter,{hout,' ',ntot,coord(0,0),ncw})
  c_proc(SetConsoleCursorPosition,{hout,coord(0,0)})
  free(ncw)
end procedure

public procedure setWindowSize(integer columns, integer lines)
  sequence cmd = sprintf("mode con LINES=%d COLS=%d", {lines, columns})
  system(cmd)
end procedure

public procedure gotoXY(integer x, integer y)
  integer hout=c_func(GetStdHandle,{STD_OUTPUT_HANDLE})
  c_proc(SetConsoleCursorPosition,{hout,coord(x-1,y-1)})
end procedure

public function whereX()
  sequence ci = getConsoleInfo()
  return ci[CURSOR_COL]
end function

public function whereY()
  sequence ci = getConsoleInfo()
  return ci[CURSOR_ROW]
end function

public function getScreenHeight()
  sequence ci = getConsoleInfo()
  return ci[BUFFER_ROWS]
end function

public function getScreenWidth()
  sequence ci = getConsoleInfo()
  return ci[BUFFER_COLS]
end function

public procedure clrEol()
  sequence ci = getConsoleInfo()
  integer nc=ci[BUFFER_COLS]-ci[CURSOR_COL]
  atom where=coord(ci[CURSOR_COL],ci[CURSOR_ROW])
  integer hout=c_func(GetStdHandle,{STD_OUTPUT_HANDLE})
  atom ncw=allocate(4)
  c_proc(FillConsoleOutputCharacter,{hout,' ',nc,where,ncw})
  c_proc(FillConsoleOutputAttribute,{hout,ci[CHAR_ATTRS],nc,where,ncw})
  free(ncw)
end procedure

public procedure insLine()
  sequence ci = getConsoleInfo()
  atom rect=allocate(8)
  poke(rect,{ci[WIN_LEFT],0,ci[WIN_TOP],0,ci[WIN_RIGHT],0,ci[WIN_BOTTOM],0})
  atom charinfo=allocate(4)
  poke(charinfo,{' ',0,ci[CHAR_ATTRS],0})
  integer hout=c_func(GetStdHandle,{STD_OUTPUT_HANDLE})
  c_proc(ScrollConsoleScreenBuffer,{hout,rect,NULL,coord(ci[WIN_LEFT],1+ci[WIN_TOP]),charinfo})
  free(rect)
  free(charinfo)
end procedure

public procedure delLine()
  sequence ci = getConsoleInfo()
  atom rect=allocate(8)
  poke(rect,{ci[WIN_LEFT],0,1+ci[WIN_TOP],0,ci[WIN_RIGHT],0,ci[WIN_BOTTOM],0})
  atom charinfo=allocate(4)
  poke(charinfo,{' ',0,ci[CHAR_ATTRS],0})
  integer hout=c_func(GetStdHandle,{STD_OUTPUT_HANDLE})
  c_proc(ScrollConsoleScreenBuffer,{hout,rect,NULL,coord(ci[WIN_LEFT],ci[WIN_TOP]),charinfo})
  free(charinfo)
  free(rect)
end procedure

public procedure getText(integer left, integer top, integer right,
                         integer bottom, atom textbuf)
  atom rect=allocate(8)
  integer hout=c_func(GetStdHandle,{STD_OUTPUT_HANDLE})
  poke(rect,{left-1,0,top-1,0,right-1,0,bottom-1,0})
  integer bufx=(1+right-left)
  integer bufy=(1+bottom-top)
  integer nbuf=bufx*bufy
  atom charinfobuf=allocate(4*nbuf)
  c_proc(ReadConsoleOutput, {hout, charinfobuf,
         coord(bufx,bufy), coord(0,0),rect})
  for i=0 to (nbuf-1) do
    sequence sdm=peek({charinfobuf+4*i,4})
    poke(textbuf+2*i,{sdm[1],sdm[3]})
  end for
  free(charinfobuf)
  free(rect)
end procedure

public procedure putText(integer left, integer top, integer right,
                         integer bottom, atom textbuf)
  atom rect=allocate(8)
  integer hout=c_func(GetStdHandle,{STD_OUTPUT_HANDLE})
  poke(rect,{left-1,0,top-1,0,right-1,0,bottom-1,0})
  integer bufx=(1+right-left)
  integer bufy=(1+bottom-top)
  integer nbuf=bufx*bufy
  atom charinfobuf=allocate(4*nbuf)
  for i=0 to (nbuf-1) do
    sequence sdm=peek({textbuf+2*i,2})
    poke(charinfobuf+4*i,{sdm[1],0,sdm[2],0})
  end for
  c_proc(WriteConsoleOutput, {hout, charinfobuf,
         coord(bufx,bufy), coord(0,0),rect})
  free(charinfobuf)
  free(rect)
end procedure

public procedure setWindowTitle(sequence s)
  integer hout=c_func(GetStdHandle,{STD_OUTPUT_HANDLE})
  atom cs=allocate_string(s)
  c_proc(SetConsoleTitle,{cs})
  free(cs)
end procedure

public function waitInput(integer timeout=-1)
  integer event = 0
  atom inrec=allocate(INREC_BYTES)
  atom nr=allocate(4)
  integer hin=c_func(GetStdHandle,{STD_INPUT_HANDLE})
  atom t0 = time()
  atom struct = 0
  sequence params = {}
  while not event do
    c_proc(ReadConsoleInput,{hin,inrec,1,nr})
    event = peek2u(inrec+INREC_EVENT)
    if (timeout >= 0) and ((time()-t0) >= timeout) then exit end if
  end while
  struct = inrec+INREC_STRUCT
  switch event do
  case KEY_EVENT then
    atom keyDown = peek4u(struct+KEY_DOWN)
    integer repeatCount = peek2u(struct+REPEAT_COUNT)
    integer virtualKeyCode = peek2u(struct+VIRTUAL_KEYCODE)
    integer virtualScanCode = peek2u(struct+VIRTUAL_SCANCODE)
    integer code = peek2u(struct+KEY_CHAR)
    atom controlKeyState = peek4u(struct+CONTROL_KEY_STATE)
    params = {keyDown, repeatCount, virtualKeyCode, virtualScanCode, code, controlKeyState}
  case MOUSE_EVENT then
    integer mouseX = peek2u(struct+MOUSE_POSITION_X)
    integer mouseY = peek2u(struct+MOUSE_POSITION_Y)
    atom buttonState = peek4u(struct+BUTTON_STATE)
    atom controlKeyState = peek4u(struct+MOUSE_CONTROL_KEYS)
    atom eventFlags = peek4u(struct+EVENT_FLAGS)
    params = {mouseX, mouseY, buttonState, controlKeyState, eventFlags}
  case WINDOW_BUFFER_SIZE_EVENT then
    integer winSizeX = peek2u(struct+WINDOW_SIZE_X)
    integer winSizeY = peek2u(struct+WINDOW_SIZE_Y)
    params = {winSizeX, winSizeY}
  end switch
  return {event, params}
end function

-- right ALT key pressed
public function isRightAltPressed(integer controlKeyState)
  return and_bits(controlKeyState, RIGHT_ALT_PRESSED)
end function

-- left ALT key pressed
public function isLeftAltPressed(integer controlKeyState)
  return and_bits(controlKeyState, LEFT_ALT_PRESSED)
end function

-- right CTRL key pressed
public function isRightCtrlPressed(integer controlKeyState)
  return and_bits(controlKeyState, RIGHT_CTRL_PRESSED)
end function

-- left CTRL key pressed
public function isLeftCtrlPressed(integer controlKeyState)
  return and_bits(controlKeyState, LEFT_CTRL_PRESSED)
end function

-- SHIFT key pressed
public function isShiftPressed(integer controlKeyState)
  return and_bits(controlKeyState, SHIFT_PRESSED)
end function

-- NUM LOCK light is on
public function isNumLockOn(integer controlKeyState)
  return and_bits(controlKeyState, NUMLOCK_ON)
end function

-- SCROLL LOCK light is on
public function isScrollLockOn(integer controlKeyState)
  return and_bits(controlKeyState, SCROLLLOCK_ON)
end function

-- CAPS LOCK light is on
public function isCapsLockOn(integer controlKeyState)
  return and_bits(controlKeyState, CAPSLOCK_ON)
end function

-- Enhanced key pressed
public function isEnhancedKeyPressed(integer controlKeyState)
  return and_bits(controlKeyState, ENHANCED_KEY)
end function

-- left Win Start pressed
public function isLeftWinStartPressed(integer controlKeyState, integer virtualKeyCode)
  return isEnhancedKeyPressed(controlKeyState) and (virtualKeyCode=91)
end function

-- right Win Start pressed
public function isRightWinStartPressed(integer controlKeyState, integer virtualKeyCode)
  return isEnhancedKeyPressed(controlKeyState) and (virtualKeyCode=92)
end function

-- Win Menu pressed
public function isWinMenuPressed(integer controlKeyState, integer virtualKeyCode)
  return isEnhancedKeyPressed(controlKeyState) and (virtualKeyCode=93)
end function

public function getCh()
  sequence event = {}
  while 1 do
    event = waitInput()
    if event[1] = KEY_EVENT then
      atom keyDown = event[2][1]
      if keyDown then exit end if
    end if
  end while
  integer virtualKeyCode = event[2][3]
  integer code = event[2][5]
  atom controlKeyState = event[2][6]
  return {code, controlKeyState, virtualKeyCode}
end function

public function getMouse()
  return 0  -- {position, buttonState, controlKeyState}
end function

public function getNewWindowSize()
  return 0  -- {cols, rows}
end function

public function cGets()
  sequence result = ""
  integer pointer = 0
  while(1) do
    sequence gs = getCh()
    integer code = gs[1]
    if code >= ' ' and code <= '~' then
      result = append(result, code)
      pointer += 1
      cPuts({code})
    elsif (pointer and code=8) then  --backspace
      pointer -= 1
      result = result[1..pointer]
      cPuts({8,' ',8})
    elsif code='\r' or code='\n' then exit
    elsif code=27 then result="" exit  --escape forces null return
    end if
  end while
  return result
end function

public function seq2int(sequence s)
  integer n, minus, c, k, issign
  sequence s1
  s1=""
  minus=0 k=0 issign=0
  for i=1 to length(s) do
    c=s[i]
    if c>='0' and c<='9' then s1=append(s1,c) k=k+1
    elsif c='-' and k=0 and issign=0 then minus=1 issign=1
    elsif c='+' and k=0 and issign=0 then issign=1
    elsif c=' ' and k=0 and issign=0 then  --ignore leading blank
    else exit  --parsing stops at unexpected character
    end if
  end for
  if k=0 then return "nan" end if
  n=0
  for i=1 to k do n=10*n+(s1[i]-'0') end for
  if minus then n= -n end if
  return n
end function

public function seq2float(sequence s)
  atom x, fp
  integer c, minus, iexp, eminus,
    issign, isd, isesign, ise, ki, kf, ke
  sequence si, sf, se
  si="" sf="" se=""
  minus=0 eminus=0 issign=0 isd=0 ise=0 isesign=0
  ki=0 kf=0 ke=0
  for i=1 to length(s) do
    c=s[i]
    if c>='0' and c<='9' then
      if ise then se=append(se,c) ke=ke+1
      elsif isd then sf=append(sf,c) kf=kf+1
      else si=append(si,c) ki=ki+1
      end if
    elsif c='-' and isd=0 and ki=0 and issign=0 then minus=1 issign=1
    elsif c='+' and isd=0 and ki=0 and issign=0 then issign=1
    elsif c='-' and ise and ke=0 and isesign=0 then eminus=1 isesign=1
    elsif c='+' and ise and ke=0 and isesign=0 then isesign=1
    elsif c='.' and isd=0 then isd=1
    elsif (c='e' or c='E') and ise=0 then ise=1
    elsif c=' ' and ki=0 and isd=0 and ise=0 and issign=0 then
      --ignore leading blank
    else exit  --parsing stops at unexpected character
    end if
  end for
  if ki=0 and kf=0 then return "nan" end if
  x=0.0
  for i=1 to ki do x=10*x+(si[i]-'0') end for
  fp=1.0
  for i=1 to kf do fp=0.1*fp x=x+fp*(sf[i]-'0') end for
  if minus then x= -x end if
  iexp=0
  for i=1 to ke do iexp=10*iexp+(se[i]-'0') end for
  if eminus then iexp= -iexp end if
  x=x*power(10,iexp)
  return x
end function

public procedure cBreakOn()
  allow_break(1)
end procedure

public procedure cBreakOff()
  allow_break(0)
end procedure
