-- myConsole.e
-- v1.1
-- 2012-09-09
-- console functions

-------------------------------------------------------------------------------

-- public sequence TextAttr
-- public integer  TextRows, TextCols
-- public constant KEYB = 0, SCREEN = 1, ERR = 2

-- public procedure setTextAttr(sequence col)
-- public function  editText(sequence s, sequence attr, sequence pos, integer longueur)
-- public procedure displayMsg(sequence msg, object pos)
-- public function  listBox(sequence liste, integer defaut, sequence attr, sequence pos, integer long)
-- public function  comboBox(sequence liste, integer defaut, sequence attr, sequence pos, integer long)

-------------------------------------------------------------------------------

include std/console.e
include std/math.e
include std/sequence.e
include std/text.e
include std/graphics.e
include myLibs/myTypes.e
include myLibs/myDebug.e
include myLibs/myConv.e

public sequence TextAttr
sequence vid_conf

public integer TextRows, TextCols
TextAttr = {WHITE, BLACK}
TextRows = 24
TextCols = 80
vid_conf = video_config()
TextRows = vid_conf[VC_LINES]
TextCols = min({80,vid_conf[VC_COLUMNS]})
--VC_SCRNLINES,
--VC_SCRNCOLS

public constant KEYB = 0, SCREEN = 1, ERR = 2

public constant
  ENTER={13}, ESC={27}, SPACE={32}, BACKSPACE = {8}, TAB = {9},
  UP = {1016400}, DOWN = {1016432}, LEFT = {1016384}, RIGHT = {1016416},
  PAGE_UP = {1016320}, PAGE_DOWN = {1016336}, HOME = {1016368},
  CTRL_LEFT = {1024176}, START_LEFT = {1020096}, ALT = {1032400}, ALT_GR = {1020096},
  START_RIGHT = {1016480}, CTXT_MENU = {1016496}, CTRL_RIGHT = {1020096},
  END = {1016352}, INSERT = {1016288}, DELETE = {1016304} 

public constant F_ = {
  {1017600}, {1017616}, {1017632}, {1017648},
  {1017664}, {1017680}, {1017696}, {1017712},
  {1017728}, {1017744}, {1017760}, {1017776}
}

integer
  listLine, listSelected, listOffset

------------------------------------------------------------------------------

public procedure setTextAttr(sequence col)
  text_color(col[1])
  bk_color(col[2])
  TextAttr = col
end procedure

------------------------------------------------------------------------------

public procedure clrEol(integer col)
-- clear end of line
  sequence CurPos, vc
  vc = video_config()
  CurPos = get_position()
  for i = CurPos[2] to vc[VC_COLUMNS] do
    put_screen_char(CurPos[1], i, {' ', col})
  end for
end procedure -- Clreol()

------------------------------------------------------------------------------

public procedure printAt(sequence pos, object obj, sequence format, sequence col)
  sequence Attr

  if sequence(obj) then
    if length(obj)=0 then return end if
  end if
  Attr = TextAttr
  setTextAttr(col)
  position(pos[1], pos[2])
  if sequence(obj) then
    printf(1, format, {obj})
  else
    printf(1, format, obj)
  end if
  setTextAttr(Attr)
end procedure

-------------------------------------------------------------------------------

public procedure displayMsg(sequence msg, object pos)
-- prints a message in a frame at position pos
-- if pos is an atom (mostly 0), then center the frame in the screen
  sequence saved, col, s
  integer h, w, x1, y1, x2, y2
  atom c

  col = TextAttr
  s = split(msg, '\n')
  h = length(s)
  w = 0
  for i = 1 to h do
    if length(s[i]) > w then w = length(s[i]) end if
  end for
  if atom(pos) then
    x1 = floor((TextCols-w)/2)
    y1 = floor((TextRows-h)/2)
    x2 = x1+w-1
    y2 = y1+h-1
  else
    x1 = pos[2]
    x2 = pos[2]+w-1
    y1 = pos[1]
    y2 = pos[1]+h-1
  end if
  saved = save_text_image({y1,x1}, {y2,x2})
  position(y1,x1)
  setTextAttr({GRAY,WHITE})
  puts(1,  msg)
  c = wait_key()
  display_text_image({y1,x1}, saved)
  setTextAttr(col)
end procedure

-------------------------------------------------------------------------------

public function getKeys()
  sequence result
  integer code

  result = {}
  code = wait_key()
  while code != -1 do
    result = append(result,code)
    code = get_key()
  end while
--  printf(f_debug, "getKeys = %s\n", {sprint(result)})
  return result
end function

------------------------------------------------------------------------------

public function editText(sequence s, sequence attr, sequence pos, integer longueur)
  sequence chaine
  integer l, pointeur_chaine
  sequence ex
  sequence col
  atom fl_out, fl_ins, premier

  col = TextAttr
  fl_ins = 1
  cursor(UNDERLINE_CURSOR)
  fl_out = 0
  setTextAttr(attr)
  chaine = s  l = 0
  pointeur_chaine = 1
  premier = 1
  while fl_out = 0 do
    position(pos[1], pos[2])  puts(1, repeat(' ', longueur))
    position(pos[1], pos[2])  puts(1, mid(chaine, 1, longueur))
    position(pos[1], pos[2] + pointeur_chaine - 1)
    ex = getKeys()
    if equal(ex, HOME) then
       pointeur_chaine = 1
    elsif equal(ex, END) then
      if length(chaine) < longueur then
        pointeur_chaine = length(chaine) + 1
      else
        pointeur_chaine = length(chaine)
      end if
    elsif equal(ex, LEFT) then
      if pointeur_chaine > 1 then
        pointeur_chaine = pointeur_chaine - 1
      end if
    elsif equal(ex, RIGHT) then
      if (pointeur_chaine < longueur) and (pointeur_chaine < length(chaine)+1) then
        pointeur_chaine = pointeur_chaine + 1
      end if
    elsif equal(ex, DELETE) then
      l = length(chaine)
      if pointeur_chaine <= l then
        if pointeur_chaine = 1 then
          chaine = chaine[2..l]
        elsif pointeur_chaine = l then
          chaine = chaine[1..l-1]
          pointeur_chaine = l-1
        else
          chaine = remove(chaine, pointeur_chaine)
        end if
      end if
    elsif equal(ex, INSERT) then
      fl_ins = not fl_ins
      if fl_ins then
        cursor(UNDERLINE_CURSOR)
      else
        cursor(BLOCK_CURSOR)
      end if
    elsif find(ex, F_ & {UP, DOWN, ENTER, TAB, ESC}) then
      fl_out = 1
    elsif equal(ex, BACKSPACE) then
      l = length(chaine)
      if pointeur_chaine > 1 then
        chaine = remove(chaine, pointeur_chaine-1)
        pointeur_chaine = pointeur_chaine-1
      end if
    elsif (ex[1] > 31) and (ex[1] < 255) then
      if premier then
        chaine = append("", ex[1])
        pointeur_chaine = 2
      elsif fl_ins and (length(chaine) < longueur) then
        chaine = insert(chaine, ex[1], pointeur_chaine)
        if (pointeur_chaine < longueur) and (pointeur_chaine < length(chaine)+1) then
          pointeur_chaine = pointeur_chaine + 1
        end if
      elsif (not fl_ins) and (pointeur_chaine <= length(chaine)) then
        chaine[pointeur_chaine] = ex[1]
        if (pointeur_chaine < longueur) and (pointeur_chaine < length(chaine)+1) then
          pointeur_chaine = pointeur_chaine + 1
        end if
      end if
    end if
    premier = 0
  end while
  setTextAttr(col)
  cursor(NO_CURSOR)
  return {chaine, ex}
end function

-------------------------------------------------------------------------------

procedure listGoFirst(sequence currentList)
  if length(currentList)=0 then return end if
  listSelected = 1
end procedure

-------------------------------------------------------------------------------

procedure listGoPrevious(sequence currentList)
  if length(currentList)=0 then return end if
  if listSelected=1 then return end if
  listSelected -= 1
end procedure

-------------------------------------------------------------------------------

procedure listGoNext(sequence currentList)
  if length(currentList)=0 then return end if
  if listSelected=length(currentList) then return end if
  listSelected += 1
end procedure

-------------------------------------------------------------------------------

procedure listGoLast(sequence currentList)
  listSelected = length(currentList)
end procedure

-------------------------------------------------------------------------------

procedure listGo(sequence currentList, integer pos)
  if pos = 0 then
    listSelected = 0
  elsif pos = 1 then
    listGoFirst(currentList)
  elsif pos = length(currentList) then
    listGoLast(currentList)
  else
    listSelected = pos
  end if
end procedure

-------------------------------------------------------------------------------

public function comboBox(sequence liste, integer defaut, sequence attr, sequence pos, integer long)
  sequence s, col
  sequence resultat

  col = TextAttr
  setTextAttr(attr)
  resultat = {0, {}}
  listSelected = defaut
  if (defaut = 0) or (length(liste)=0) then
    s = ""
  else
    s = liste[listSelected]
  end if
--    puts(1,  pad_tail(s[1..min({long, length(s)})], long))
  while not find(resultat[2], {ENTER,ESC}) do
    resultat = editText(s, attr, pos, long)
	puts(f_debug, sprint(resultat) & "\n")
    if equal(resultat[2],ENTER) then
	  exit
    elsif equal(resultat[2],UP) then
      listGoPrevious(liste)
      s = liste[listSelected]
    elsif equal(resultat[2],DOWN) then
      listGoNext(liste)
      s = liste[listSelected]
    elsif equal(resultat[2],ESC) then
      resultat[1] = ""
    end if
  end while
  setTextAttr(col)
  position(pos[1],pos[2])
  puts(1,  pad_tail(resultat[1], long))
  return resultat
end function
